class Phrogz::DependencyWalker
  # +dependencies+ are specified as a Hash mapping one object to the object(s)
  # that it depends upon. Objects that should be included in the traversal but
  # that do not depend on any others may be mapped to nil.
  #
  # All objects will be walked AFTER the objects they depend upon.
  # Circular references are not detected, and will cause a SystemStackError.
  # Any self-references (e.g. :a=>:a ) will hang the algorithm.
  #
  # Returns an array of objects in the traversal order.
  # If a block is supplied, it is executed on the objects in traversal order.
  #
  # Examples:
  #  include Phrogz
  #  p DependencyWalker.walk( :a=>:b )
  #  #=> [:b, :a]
  #
  #  p DependencyWalker.walk( :a=>:b, :c=>:a, :b=>[:d,:e], :e=>:d, :f=>nil )
  #  #=> [:f, :d, :e, :b, :a, :c]
  #
  #  DependencyWalker.walk( :a=>[:b], :b=>[:d,:c], :c=>:d ) do |obj|
  #    p obj
  #  end
  #  #=> :d
  #  #=> :c
  #  #=> :b
  #  #=> :a
  def self.walk( dependencies={} )
   chain = []
   # TODO: detect circular references and bail quietly
   # http://phrogz.net/nodes/traversingdirectedgraph.asp
   process = lambda do |key|
     chain << key
     dependent = dependencies[ key ]
     unless dependent.nil?
       if dependent.is_a?( Array )
         dependent.each{ |item| process[ item ] }
       else
         process[ dependent ]
       end
     end
   end
   dependencies.each{ |key,_| process[ key ] }
   chain.reverse!.uniq!
   chain.each{ |item| yield item } if block_given?
   chain
  end
  
  def self.moonwalk( dependencies={} )
    chain = self.walk( dependencies ).reverse
    chain.each{ |item| yield item } if block_given?
    chain
  end
  
  def initialize( dependencies = {} )
    @dependencies = dependencies
  end
  
  def []( slave )
    @dependencies[ slave ]
  end
  
  def []=( slave, masters )
    @dependencies[ slave ] = masters
  end
  
  def walk( &blk )
    self.class.walk( @dependencies, &blk )
  end
  
  def moonwalk( &blk )
    self.class.moonwalk( @dependencies, &blk )
  end
  
  def to_s
    self.class.walk( @dependencies ).join( ' <- ' )
  end
end

