require 'RoughCSV'
require 'DependencyWalker'

module Phrogz; end
module Phrogz::TableTransloader
  @transloadable_models = []
  class << self
    attr_reader :transloadable_models
  end

  attr_reader   :pk_after_transload
  attr_accessor :original_table_name, :temporary_table_name

  def transload_from( csv_file, original_pk=:id, &transformation )
    @transload_source   = csv_file
    @pk_after_transload = {}
    @pre_transload_pk   = original_pk
    @transformation     = transformation
    ::Phrogz::TableTransloader.transloadable_models << self
  end

  def transload_old_data
    rows_loaded = 0
    Phrogz::RoughCSV.load( @transload_source ).each do |old_row|
      orig_pk = old_row[ @pre_transload_pk ]
      raise "No primary key #{@pre_transload_pk} supplied for #{old_row.inspect} (#{self})" if orig_pk.nil?

      existing_row = @pk_after_transload[ orig_pk ]
      raise "Record with primary key #{orig_pk} already existed" if existing_row

      mapping = @transformation.call( old_row )
      if mapping
        # Assumes that the DB adapter always returns the primary key value
        # even when explicitly inserting a record with a primary key.
        # (SQLite3 doesn't do this as of this writing.)
        @pk_after_transload[ orig_pk ] = ( self << mapping )
        #puts "@pk_after_transload[ #{orig_pk.inspect} ] = #{@pk_after_transload[ orig_pk ]}"

        # Alternative, slower code for 'broken' DB adapters:
        # self << mapping
        # @pk_after_transload[ orig_pk ] = first(mapping).pk
      end
      rows_loaded += 1
    end
    rows_loaded
  end

  def model_references
    references = {}
    self.all_association_reflections.reject{ |assoc| assoc[:type]==:one_to_many }.each do |assoc|
      model = Object.const_get( assoc[ :class_name ] )
      references[ assoc[:key] ] = model
    end
    references
  end

  def referenced_by
    references = {}
    self.all_association_reflections.select{ |assoc| assoc[:type]==:one_to_many }.each do |assoc|
      model = Object.const_get( assoc[ :class_name ] )
      references[ model ] ||= []
      references[ model ] << assoc[ :key ]
    end
    references
  end

  def self.transload_all( *models )
    models = @transloadable_models if models.empty?
    DB.transaction do
      # All data goes into temporary tables at first where invalid foreign keys are accepted
      puts "Creating temp tables..." if $SHOW_PROGRESS
      models.each do |model|
        model.original_table_name  = model.table_name
        model.temporary_table_name = :"#{model.original_table_name}_staging"
        DB.execute <<-ENDSQL
          DROP TABLE IF EXISTS #{model.temporary_table_name};
          CREATE TABLE #{model.temporary_table_name} (
            LIKE #{model.original_table_name}
            INCLUDING DEFAULTS
            INCLUDING INDEXES
            EXCLUDING CONSTRAINTS
          );
        ENDSQL
        model.set_dataset( model.temporary_table_name )
      end

      # Shove the values in, with foreign keys still referencing old primary keys
      models.each do |model|
        puts "Loading data for #{model} into #{model.table_name}..." if $SHOW_PROGRESS
        rows_loaded = model.transload_old_data
        puts "...#{rows_loaded} row#{:s unless rows_loaded==1} loaded" if $SHOW_PROGRESS
      end

      # Calculate a comprehensive references map for each model, including references only specified by other models.
      # Hash is indexed by source model, yielding a hash of the columns mapping to referenced models.
      puts "Calculating reference map..." if $SHOW_PROGRESS
      references = {}
      models.each do |referencing_model|
        references[ referencing_model ] = referencing_model.model_references
      end
      models.each do |referenced_model|
        referenced_model.referenced_by.each do |referencing_model,dependent_columns|
          dependent_columns.each do |dependent_column|
            references[ referencing_model ][ dependent_column ] = referenced_model
          end
        end
      end
      references.each{ |source, dependents| puts "   #{source}=>#{dependents.inspect}" } if $SHOW_PROGRESS

      # Update foreign keys in all tables to point to new primary key values
      references.each do |referencing_model, dependencies|
        unless dependencies.empty?
          puts "Updating external references for #{referencing_model} in #{referencing_model.table_name}..." if $SHOW_PROGRESS
          referencing_model.each do |new_row|
            any_value_changed = false
            dependencies.each do |column,referenced_model|
              #puts "   ..updating for #{referenced_model} (#{referenced_model.table_name})"
              new_value = referenced_model.pk_after_transload[ new_row[ column ] ]
              old_value = new_row[ column ]
              if new_value != old_value
                puts "   ....#{column} #{old_value.inspect}=>#{new_value.inspect}"
                new_row[ column ] = new_value
                any_value_changed = true
              end
            end
            new_row.save if any_value_changed
          end
        end
      end

      # Find table-level dependencies
      walker = ::Phrogz::DependencyWalker.new
      references.each do |referencing_model, reference_hash|
        walker[ referencing_model ] = reference_hash.values.uniq - [referencing_model]
      end
      # Ensure that all models are accounted for during walking,
      # in case they're not part of any relationship
      models.each{ |model| walker[ model ] ||= nil }

      puts "Table dependency chain: #{walker}" if $SHOW_PROGRESS
      
      # Clear out the tables in reverse-dependent order
      walker.moonwalk do |model|
        DB.execute "DELETE FROM #{model.original_table_name};"
      end

      # Push data from temp tables back to the originals, in dependent order
      # TODO: need to apply dependent order on a per-row basis for self-referencing tables
      walker.walk do |model|
        puts "Moving data for #{model} from #{model.temporary_table_name} to #{model.original_table_name}..." if $SHOW_PROGRESS
        DB.execute <<-ENDSQL
          INSERT INTO #{model.original_table_name} SELECT * FROM #{model.temporary_table_name};
          DROP TABLE #{model.temporary_table_name};
        ENDSQL
        model.set_dataset( model.original_table_name )
      end
    end
  end

end
